# -*- coding: utf-8 -*-
"""A small library for processing images of agricultural fields"""

__version__ = '0.0.1'
__author__ = 'Raikel Bordon'
__contact__ = 'raikelbl@gmail.com'
__license__ = 'MIT'

from . import rows, veget, weeds, utils

__all__ = ('rows', 'veget', 'weeds', 'utils')
